

/* Written by Michael E. Kounavis based on the NIST cryptographic API, 
   Commmunications Technology Lab, Intel Corporation, 2008
 */


#include "SHA3api_ref.h"
#include "vortex_core.h"
#include "vortex_misc.h"

extern uint8_t number_of_rounds_g;
extern uint8_t mul_type_g;
extern uint8_t degree_of_diffusion_g;
extern uint8_t *a0_b0_32_g;
extern uint8_t *ta_tb_32_g;
extern uint8_t *a0_b0_64_g;
extern uint8_t *ta_tb_64_g;
extern uint8_t *rcon_256_g[MAX_NUMBER_OF_ROUNDS];
extern uint8_t *rcon_512_g[MAX_NUMBER_OF_ROUNDS];

#pragma warning( disable : 4013 4312)

int				iteration_g = 0;
BitSequence		*buf_g;


uint8_t
reflect(uint8_t val)
{
	uint8_t r = 0, b[8];
	int i;
	for(i=0; i < 8; i++)
		b[i] = (val >> i) & 1;
	for(i=0; i < 8; i++)
		r |= (b[7-i] << i);
	return r;
}

#ifdef HAMMING_W_ANALYSIS
uint8_t
get_hamming_w(uint8_t val)
{
	uint8_t r = 0;
	int i;
	for(i=0; i < 8; i++)
		r += ((val >> i) & 1);
	return r;
}

uint32_t
compute_hamming_weight(BitSequence *data, 
					  int length)
{
	uint32_t r=0, i, bytelen = (length+7)/8;
	for(i=0; i < bytelen; i++)
		r += get_hamming_w(data[i]);
	return r;
	
}

void
print_hamming_weight(BitSequence *data, 
					 int length)
{
	int r=0, i, bytelen = (length+7)/8;
	for(i=0; i < bytelen; i++)
		r += get_hamming_w(data[i]);
	printf("Hamming Weight %d found as output of vortex %d\n", r, length);
}
#endif

/* reflects the bits inside each byte according to the NIST specified format */

int
set_format(BitSequence *data, 
		   DataLength databitlen)
{
	uint64_t bytelen = (databitlen+7)/8;
	uint64_t i;
	for(i=0; i < bytelen; i++)
		data[i] = reflect(data[i]);
	return SUCCESS;
}

/* Note: the only variable of the hash state which is not set by the Init() is the padding length. This is
   because padding-length is message-dependent. Vortex-specific parameter values are obtained from the 
   global variables defined in vortex_core.c
 */

HashReturn Init(hashState *state, 
				int hashbitlen)
{
	set_default_parameters();
	generate_constants();
	if(!state)
	{
		perror("Init(): null pointer passed\n");
		return NULL_POINTER_ERROR;
	}
	state->hashbitlen = hashbitlen;
	state->number_of_rounds = number_of_rounds_g;
	state->mul_type = mul_type_g;
	state->degree_of_diffusion = degree_of_diffusion_g;
	state->padding_length = 0;

	switch(hashbitlen)
	{
		case 224: 
		case 256:
			state->hash  = (uint8_t *)malloc(32);
			state->a0_b0 = (uint8_t *)malloc(32);
			state->ta_tb = (uint8_t *)malloc(32);
			varcpy(state->hash,  a0_b0_32_g, 32);
			varcpy(state->a0_b0, a0_b0_32_g, 32);
			varcpy(state->ta_tb, ta_tb_32_g, 32);
			break;
		case 384: 
		case 512:
			state->hash  = (uint8_t *)malloc(64);
			state->a0_b0 = (uint8_t *)malloc(64);
			state->ta_tb = (uint8_t *)malloc(64);
			varcpy(state->hash,  a0_b0_64_g, 64);
			varcpy(state->a0_b0, a0_b0_64_g, 64);
			varcpy(state->ta_tb, ta_tb_64_g, 64);
			break;
		default:
			perror("Init(): bad hash type\n");
			return BAD_HASHBITLEN;
	}
	return SUCCESS;
}

/* Update() assumes that the data length passed is at least equal to the block size 
   of the vortex algorithm invoked. The block size is 512 bits for vortex 224 and
   vortex 256 and 1024 bits for vortex 384 and vortex 512.
 */

HashReturn Update(hashState *state, 
				  const BitSequence *data, 
				  DataLength databitlen)
{
#ifndef OPTIMIZED_ASSEMBLY
	int i;
#endif
	uint64_t blocks_no_malloc;
	if((!state) || (!data))
	{
		perror("Update(): null pointer passed\n");
		return NULL_POINTER_ERROR;
	}

	if(((state->hashbitlen == 384) || (state->hashbitlen == 512)) &&
		(databitlen == 512) &&
		(iteration_g == 0))
	{
		buf_g = (BitSequence *)malloc(128);
		varcpy(buf_g, (uint8_t *)data, (uint32_t)64);
		state->padding_length += databitlen;
		iteration_g = 1;
		return SUCCESS;
	}
	else if (((state->hashbitlen == 384) || (state->hashbitlen == 512)) &&
			(databitlen == 512) &&
			(iteration_g == 1))
	{
		varcpy(buf_g+64, (uint8_t *)data, (uint32_t)64);
		set_format(buf_g, 1024);
		iteration_g = 0;
	}
	else
	{
		buf_g = (BitSequence *)malloc((databitlen+7)/8);
		varcpy(buf_g, (uint8_t *)data, (uint32_t)((databitlen+7)/8));
		set_format(buf_g, databitlen);
	}
	switch(state->hashbitlen)
	{
		case 224: 
		case 256:
			/* update processes multiples of the block size neglecting the remaining bits 
			   as specified by NIST */
			blocks_no_malloc = databitlen/512;
#ifdef OPTIMIZED_ASSEMBLY
		vortex_256_asm((unsigned char *)state->hash, (unsigned char *)buf_g, (unsigned char *)rcon_256_g, (int)blocks_no_malloc);
#else
		for(i=0; i< blocks_no_malloc; i++)
			vortex_256_block(state->hash, &buf_g[i*64], state->number_of_rounds, state->mul_type);
#endif
			/* 
			   change made to the implementation of the API 
			   so that genKAT.c does not require any changes
			*/
			state->padding_length += databitlen;
			break;
		case 384: 
		case 512:
			if(databitlen == 512)
				blocks_no_malloc = 1;
			else
				blocks_no_malloc = databitlen/1024;
#ifdef OPTIMIZED_ASSEMBLY
		vortex_512_asm((unsigned char *)state->hash, (unsigned char *)buf_g, (unsigned char *)rcon_512_g, (int)blocks_no_malloc);
#else
			for(i=0; i< blocks_no_malloc; i++)
				vortex_512_block(state->hash, &buf_g[i*128], state->number_of_rounds, state->mul_type);
#endif
			/* 
			   change made to the implementation of the API 
			   so that genKAT.c does not require any changes
			*/
			state->padding_length += databitlen;
			break;
		default:
			perror("Init(): bad hash type\n");
			return BAD_HASHBITLEN;
	}
	free(buf_g);
	return SUCCESS;
}

/* Final() implements the last block of the Enveloped Merkle-Damgard construction
 */

HashReturn Final(hashState *state, 
				 BitSequence *hashval)
{
	uint8_t last_block[128];
	varset(last_block, 0, 128);

	if((!hashval) || (!state))
	{
		perror("Final(): null pointer passed\n");
		return NULL_POINTER_ERROR;
	}
	switch(state->hashbitlen)
	{
		case 224: 
		case 256:	
			varcpy(last_block, state->hash, 32);
			last_block[32] = 1;
			*(uint64_t *)&last_block[56] = state->padding_length;
			varcpy(state->hash, state->ta_tb, 32);
			last_vortex_256_block(state->hash, last_block, state->number_of_rounds, state->mul_type, state->degree_of_diffusion);
			varcpy((uint8_t *)hashval, state->hash, 32);
			break;
		case 384:
		case 512:
			varcpy(last_block, state->hash, 64);
			last_block[64] = 1;
			*(uint64_t *)&last_block[120] = state->padding_length;
			varcpy(state->hash, state->ta_tb, 64);
			last_vortex_512_block(state->hash, last_block, state->number_of_rounds, state->mul_type, state->degree_of_diffusion);
			varcpy((uint8_t *)hashval, state->hash, 64);
			break;
		default:
			perror("Final(): bad hash type\n");
			return BAD_HASHBITLEN;
	}
	set_format(hashval, (DataLength)state->hashbitlen);
	return SUCCESS;
}


HashReturn Hash(int hashbitlen, 
				const BitSequence *data, 
				DataLength databitlen, 
				BitSequence *hashval)
{
	int error_code;
	uint64_t l[2];
	BitSequence *buf;
	l[1] = 0;
	l[0] = databitlen;
	if((!hashval) || (!data))
	{
		perror("Hash(): null pointer passed\n");
		return NULL_POINTER_ERROR;
	}
	set_default_parameters();
	generate_constants();
	buf = (BitSequence *)malloc((databitlen+7)/8);
	varcpy(buf, (uint8_t *)data, (uint32_t)((databitlen+7)/8));
	set_format((BitSequence *)buf, databitlen);
	switch(hashbitlen)
	{
		case 224: 
		case 256:
			error_code = tunable_vortex(buf, l, (uint8_t *)hashval, (uint32_t)hashbitlen,
				number_of_rounds_g, mul_type_g, a0_b0_32_g, ta_tb_32_g, degree_of_diffusion_g);
			break;
		case 384:
		case 512:
			error_code = tunable_vortex(buf, l, (uint8_t *)hashval, (uint32_t)hashbitlen,
				number_of_rounds_g, mul_type_g, a0_b0_64_g, ta_tb_64_g, degree_of_diffusion_g);
			break;
		default:
			perror("Hash(): bad hash type\n");
			return BAD_HASHBITLEN;
	}
	if(error_code == SUCCESS)
	{
		set_format(hashval, (DataLength)hashbitlen);
	}
	free(buf);
	return error_code;
}
